from __future__ import annotations

import json
import typing
from base64 import b64decode

import itsdangerous
from itsdangerous.exc import BadSignature
from starlette.datastructures import Secret
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Receive, Scope, Send

from . import config
from .globals import g


class MessageFlashMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        secret_key: str | Secret = config.Config.SECRET_KEY,
        path: str = "/",
        same_site: typing.Literal["lax", "strict", "none"] = "lax",
        https_only: bool = False,
        domain: str | None = None,
    ) -> None:
        self.app = app
        self.signer = itsdangerous.TimestampSigner(str(secret_key))
        self.message_flash_cookie = config.Config.MESSAGE_FLASH_COOKIE
        self.max_age = 60 * 60  # 1 hr
        self.path = path
        self.security_flags = "httponly; samesite=" + same_site
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; secure"
        if domain is not None:
            self.security_flags += f"; domain={domain}"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)

        # Read flash messages if the cookie exists and set g.flash_messages
        if self.message_flash_cookie in connection.cookies:
            data = connection.cookies[self.message_flash_cookie].encode("utf-8")

            try:
                data = self.signer.unsign(data, max_age=self.max_age)
                g.flash_messages = json.loads(b64decode(data))
            except BadSignature:
                g.flash_messages = []
        else:
            g.flash_messages = []

        await self.app(scope, receive, send)



import pytest

from mojito.testclient import TestClient

from .main import app

client = TestClient(app)


@pytest.mark.parametrize(
    "path,status,expected_response",
    [
        ("/", 200, "index_response"),
        ("/doesnt_exist", 404, "Not Found"),
        ("/async_route", 200, "async_route_response"),
    ],
)
def test_get_index(path: str, status: int, expected_response: str):
    response = client.get(path)
    assert response.status_code == status
    assert response.text == expected_response


@pytest.mark.parametrize(
    "path_param,query_param_1,status", [(1, "param", 200), ("nope", "param", 404)]
)
def test_params(path_param: int, query_param_1: str, status: int):
    response = client.get(f"/{path_param}?query_param_1={query_param_1}")
    assert response.status_code == status
    if response.status_code == 200:
        assert response.json() == {"id": path_param, "query_param_1": query_param_1}


def test_application_route():
    response = client.get("/app-route")
    assert response.status_code == 200
    assert response.text == "app-route"



from contextlib import asynccontextmanager
from typing import Any, TypeVar

from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from starlette.datastructures import UploadFile as StarletteUploadFile

from .requests import Request

try:
    from pydantic import BaseModel
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Form requires pydantic being installed. \npip install pydantic"
    )


PydanticModel = TypeVar("PydanticModel", bound=BaseModel)


@asynccontextmanager
async def Form(request: Request, model: type[PydanticModel]):
    """Read form data from the request and validate it's content against a Pydantic model
    and return the valid Pydantic model. Extra data in the form is ignored and not passed into the
    Pydantic model. This does not work for processing files. You must use the request directly to get and read
    from files before using this function to read and validate the other form fields. See
    https://www.starlette.io/requests/#request-files for working with files.

    Args:
        request (Request): Mojito Request object
        model (PydanticModel): The Pydantic model to validate against

    Returns:
        PydanticModel: The validated Pydantic model

    Raises:
        ValidationError: Pydantic validation error
    """
    async with request.form() as form:
        valid_model = model.model_validate(dict(form.items()))
        yield valid_model  # Yield result while in context of request.form()


class UploadFile(StarletteUploadFile):
    """An uploaded file included as part of the request data.

    This is a subclass of starlette.datastructures.UploadFile that can be used in a Pydantic
    BaseModel class. Pydantic will pass the model through as-is without validation.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        # Allow this file type to pass through pydantic without schema validation
        return core_schema.any_schema()



from mojito import AppRouter, Mojito, Request, auth
from mojito.testclient import TestClient

from .main import PasswordAuth

app = Mojito()
protected_router = AppRouter()
protected_router.add_middleware(auth.AuthRequiredMiddleware)

client = TestClient(app)
auth.set_auth_handler(PasswordAuth)


@protected_router.route("/login", methods=["GET", "POST"])
async def protected_login_route(request: Request):
    if request.method == "POST":
        await auth.password_login("username", "password")
        return "login success"
    return "login page"


@protected_router.route("/protected")
def protected_route():
    return "accessed"


app.include_router(protected_router)


def test_route_protection():
    result = client.get("/protected")
    assert result.status_code == 200  # Redirects to login page
    assert result.text != "accessed"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/protected")
    assert result.status_code == 200
    assert result.text == "accessed"


scope_protected_router = AppRouter()
scope_protected_router.add_middleware(
    auth.AuthRequiredMiddleware, require_scopes=["admin"]
)


@scope_protected_router.route("/scope_protected_admin")
async def scope_protected_admin():
    return "scope protected admin"


invalid_scope_protected_router = AppRouter()
invalid_scope_protected_router.add_middleware(
    auth.AuthRequiredMiddleware, require_scopes=["nope"]
)


@invalid_scope_protected_router.route("/invalid_scope_protected")
async def invalid_scope_protected():
    return "Invalid route"


app.include_router(scope_protected_router)
app.include_router(invalid_scope_protected_router)


def test_valid_scope_protected_router():
    client.cookies.clear()
    result = client.get("/scope_protected_admin")
    assert result.status_code == 200
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/scope_protected_admin")
    assert result.status_code == 200
    assert result.text == "scope protected admin"


def test_invalid_scope_protected_router():
    client.cookies.clear()
    result = client.get("/invalid_scope_protected")
    assert result.status_code == 200
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/invalid_scope_protected")
    assert result.status_code == 200
    assert result.text == "login page"


@app.route("/decorator_protected")
@auth.require_auth()
def decorator_protected_route():
    return "decorator protected"


def test_decorator_protected():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected")
    assert result.status_code == 200
    assert result.text == "decorator protected"


@app.route("/decorator_protected_with_scope")
@auth.require_auth(scopes=["admin"])
def decorator_protected_with_scopes():
    return "decorator protected with scope"


@app.route("/decorator_protected_missing_scope")
@auth.require_auth(scopes=["nope"])
def decorator_protected_missing_scope():
    return "decorator protected missing scope"


def test_decorator_protected_with_scope():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected_with_scope")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected with scope"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected_with_scope")
    assert result.text == "decorator protected with scope"


def test_decorator_protected_missing_scope():
    client.cookies.clear()  # Clear cookies
    result = client.get("/decorator_protected_missing_scope")
    assert result.status_code == 200  # Redirect to login page
    assert result.text != "decorator protected missing scope"
    assert result.text == "login page"
    result = client.post("/login")
    assert result.status_code == 200
    result = client.get("/decorator_protected_missing_scope")
    assert result.text != "decorator protected missing scope"
    assert result.text == "login page"



from asyncio import sleep

from mojito import (
    AppRouter,
    Jinja2Templates,
    JSONResponse,
    Mojito,
    Request,
    auth,
)

app = Mojito()

templates = Jinja2Templates("tests/templates")

main_router = AppRouter()


@main_router.route("/")
async def index():
    return "index_response"


@main_router.route("/async_route")
async def async_route():
    return "async_route_response"


@main_router.route("/{id:int}")
async def id_route_with_query_params(id: int, query_param_1: str, request: Request):
    return JSONResponse({"id": id, "query_param_1": query_param_1})


# TEST PROTECTED ROUTES
protected_subrouter = AppRouter("/protected")
protected_subrouter.add_middleware(auth.AuthRequiredMiddleware)


@protected_subrouter.route("/")
def protected_route():
    return "<p>protected</p>"


class PasswordAuth(auth.BaseAuth):
    async def authorize(self, request: Request, scopes: list[str]) -> bool:
        await sleep(0.5)
        return True

    async def authenticate(self, request: Request, username: str, password: str):
        await sleep(0.5)
        return (True, ["admin"])


@app.route("/app-route")
def app_route():
    return "app-route"


main_router.include_router(protected_subrouter)

auth.set_auth_handler(PasswordAuth)

app.include_router(main_router)

if __name__ == "__main__":
    import uvicorn

    for route in app.routes:
        print(f"route: {route}")
    uvicorn.run("tests.main:app", reload=True)



import json
import typing as t
from base64 import b64encode

import itsdangerous

from .config import Config
from .globals import g


class MessageFlash(t.TypedDict):
    message: str
    category: t.Optional[str]


def encode_message_cookie(message: list[MessageFlash]) -> bytes:
    data = b64encode(json.dumps(g.next_flash_messages).encode("utf-8"))
    cookie = itsdangerous.TimestampSigner(str(Config.SECRET_KEY)).sign(data)
    return cookie


def flash_message(message: str, category: t.Optional[str] = None):
    if not g.next_flash_messages:
        g.next_flash_messages = []
    message_flash = MessageFlash(message=message, category=category)
    g.next_flash_messages.append(message_flash)  # type:ignore


def get_flashed_messages() -> t.Optional[list[MessageFlash]]:
    return g.flash_messages



# TODO - Load config from os if it exists
class Config:
    SECRET_KEY: str = "poaiuvnepwoijsoxmkcpaoiu"
    MESSAGE_FLASH_COOKIE: str = "flash_messages"
    LOGIN_URL: str = "/login"
    SESSION_EXPIRES: int = 60 * 60 * 24 * 7  # 1wk default
    "In seconds. How long the session cookie should be allowed to exist before being replaced."
    SESSION_REVALIDATE_AFTER: int = 0  # Revalidate on every request
    """In seconds.
    How long the session cookie should be considered valid before running revalidation
      on the users authentication. Within this time, as long as the cookie signature is
      valid, the user will be considered authenticated and all the cookie data will be used.
      """



import os

import pytest
from pydantic import (
    BaseModel,
    Field,
    ValidationError,
)

from mojito import Form, JSONResponse, Mojito, Request, Response
from mojito.form import UploadFile
from mojito.testclient import TestClient

app = Mojito()
client = TestClient(app)


class FormTest(BaseModel):
    field_1: str
    field_2: str
    field_3: int = Field(default=10)  # Optional field


@app.route("/", methods=["POST"])
async def process_form(request: Request):
    try:
        async with Form(request, FormTest) as form:
            return JSONResponse(form.model_dump())
    except ValidationError as e:
        return Response(e.__str__(), status_code=500)


@pytest.mark.parametrize(
    ("form_data", "status"),
    [
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "field_3": 15,
            },
            200,
        ),
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "field_3": "nope",  # Invalid type
            },
            500,
        ),
        ({"field_1": "field one", "field_2": "field two"}, 200),
        (
            {
                "field_1": "field one",
                "field_2": "field two",
                "extra_field": "extra field",  # Ignored field
            },
            200,
        ),
        (
            {
                "field_1": "field one",  # Missing field_2
            },
            500,
        ),
    ],
)
def test_form_processing(form_data: dict[str, str], status: int):
    result = client.post("/", data=form_data)
    assert result.status_code == status
    if result.status_code == 200:
        assert result.json().get("field_1") == form_data.get("field_1")


class FormWithFileTest(BaseModel):
    field_1: str = ""
    file_1: UploadFile


@app.route("/upload_file", methods=["POST"])
async def process_upload_file_form(request: Request):
    async with Form(request, FormWithFileTest) as form:
        file_content = await form.file_1.read()
        return file_content.decode()


def test_form_file_processing():
    directory = os.path.dirname(__file__)
    file = open(directory + "/txt_file.txt", "rb")
    file_content = file.read()
    result = client.post("/upload_file", files={"file_1": file})
    assert result.status_code == 200
    assert result.text == file_content.decode()



"Authentication and authorization handlers built off of sessions and scoped to the request to allow for database lookups"

# A common need is to validate a session cookie and perform a database lookup to validate
# the user. This is not a middlware so that the current request can be included in the handler functions
import abc
import functools
import hashlib
import typing as t

from starlette.requests import Request
from starlette.responses import RedirectResponse
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from .config import Config
from .globals import g


class _RequestSession:
    "Names of data set on the request.session"

    IS_AUTHENTICATED = "is_authenticated"
    AUTH_SCOPES = "auth_scopes"


class AuthRequiredMiddleware:
    """Redirect to login_url if session is not authenticated or if user does not have the required auth scopes.
    Can be applied at the app level or on individual routers.

    Will ignore the Config.LOGIN_URL path to prevent infinite redirects.

    Args:
        ignore_routes (Optional[list[str]]): defaults to None. paths of routes to ignore validation on like '/login'. Path should be relative
            and match the Request.url.path value when the route is called.
        required_scopes (Optional[list[str]]) defaults to None. List of scopes the user must have in order to be authorized
            to access the requested resource.
    """

    def __init__(
        self,
        app: ASGIApp,
        ignore_routes: list[str] = [],
        require_scopes: t.Optional[list[str]] = None,
    ) -> None:
        self.app = app
        self.ignore_routes = ignore_routes
        self.require_scopes = require_scopes

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            return await self.app(scope, receive, send)
        request = Request(scope, receive)

        async def send_wrapper(message: Message) -> None:
            # ... Do something
            if (
                request.url.path in self.ignore_routes
                or request.url.path == Config.LOGIN_URL
            ):
                # Skip for routes registered as login_not_required
                return await send(message)
            is_authenticated: t.Optional[bool] = request.session.get(
                _RequestSession.IS_AUTHENTICATED
            )
            if not is_authenticated:
                response = RedirectResponse(Config.LOGIN_URL, 302)
                return await response(scope, receive, send)
            # Check that the user has the required scopes
            user_scopes = request.session.get(_RequestSession.AUTH_SCOPES, [])
            for required_scope in self.require_scopes if self.require_scopes else []:
                if not required_scope in user_scopes:
                    response = RedirectResponse(Config.LOGIN_URL, 302)
                    return await response(scope, receive, send)
            await send(message)

        await self.app(scope, receive, send_wrapper)


def require_auth(
    scopes: t.Optional[list[str]] = None, redirect_url: t.Optional[str] = None
):
    """Decorator to require that the user is authenticated and optionally check that the user has
    the required auth scopes before accessing the resource. Redirect to the configured
    login_url if one is set, or to redirect_url if one is given.

    Args:
        scopes (Optional[list[str]]): Auth scopes to verify the user has. Defaults to None.
        redirect_url (Optional[list[str]]): Redirect to this url rather than the configured
            login_url.
    """
    # This decorator must be applied below the route definition decorator so that it will wrap the
    # endpoint function before the route decorator will. This decorator will pass all arg straight
    # through after verifying authentication or else it will return a redirect response.

    def wrapper(func: t.Callable[..., t.Any]):
        @functools.wraps(func)
        def requires_auth_function(*args: t.Any, **kwargs: dict[str, t.Any]):
            request = g.request
            assert isinstance(request, Request)
            REDIRECT_URL = redirect_url if redirect_url else Config.LOGIN_URL
            if not request.session.get(_RequestSession.IS_AUTHENTICATED, False):
                return RedirectResponse(REDIRECT_URL, 302)
            if scopes:
                user_scopes = request.session.get(_RequestSession.AUTH_SCOPES, [])
                for required_scope in scopes:
                    if not required_scope in user_scopes:
                        return RedirectResponse(REDIRECT_URL, 302)
            return func(*args, **kwargs)

        return requires_auth_function

    return wrapper


class BaseAuth:
    """Base class that all authentication methods should implement.

    Subclasses must implement authorize() and authenticate() methods.
    """

    @abc.abstractmethod
    async def authorize(self, request: Request, scopes: list[str]) -> bool:
        """Method to check if the user has the required scopes. The user must have all
        scopes given to be valid.

        Args:
            scopes (list[str]): list of scopes to check check that the user has.

        Raises:
            NotImplementedError: Method not implemented.

        Returns:
            bool: User is authorized
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def authenticate(
        self, request: Request, username: str, password: str
    ) -> t.Optional[tuple[bool, list[str]]]:
        """Method to authenticate the user based on the users username and password. Will
        be used by the password_login() function to authenticate the user.

        Args:
            request (Request): Mojito/Starlette request object
            username (str): The users username
            password (str): The users password in plain text. Stored passwords should be
                hashed and compared to check validity. This base class provides hash_password()
                static method to easily compare the hashed vs the given password.

        Raises:
            NotImplementedError: Method not implemented

        Returns:
            tuple[bool, list[str]] | None: (is_authenticated: bool, user_scopes: list[str])
        """
        raise NotImplementedError()


def hash_password(password: str) -> str:
    """Helper to hash a password before storing it or to compare a plain text password to the one stored.

    Args:
        password (str): The plain text password to hash.

    Returns:
        str: The hashed password.
    """
    return hashlib.sha256(password.encode()).hexdigest()


class AuthConfig:
    """Global configuration options for auth functionality."""

    auth_handler: t.Optional[type[BaseAuth]] = None


def set_auth_handler(handler: type[BaseAuth]):
    """Set and auth handler to the AuthConfig.auth_handler setting. Only one auth handler can
    be set at a time. Setting an auth handler will override any set previously.

    Args:
        handler (type[BaseAuth]): The auth handler class that implements the BaseAuth class
    """
    AuthConfig.auth_handler = handler


async def password_login(username: str, password: str):
    """Login user based on username and password. Authenticates user and sets data on the
    session object. Uses the

    Args:
        username (str): The username to identify the user
        password (str): The users plain text password. Will be compared to the hashed version in storage.
    """
    request: Request = g.request
    if not AuthConfig.auth_handler:
        raise NotImplementedError(
            "an auth handler must be defined to use password_login"
        )
    auth = AuthConfig.auth_handler()  # Get Auth class from config
    result = await auth.authenticate(
        request=request, username=username, password=password
    )
    if not result:
        return False
    is_authenticated = result[0]
    request.session[_RequestSession.IS_AUTHENTICATED] = is_authenticated
    request.session[_RequestSession.AUTH_SCOPES] = result[1]
    return True if is_authenticated else False
